@@ -3,12 +3,15 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveGeneric #-}
 {-# OPTIONS_GHC -fno-warn-missing-fields #-}
 {-# OPTIONS_GHC -fno-warn-incomplete-patterns #-}
 -- | For lack of a better name... a parameterized version of Julius.
 module Text.Shakespeare
     ( ShakespeareSettings (..)
     , PreConvert (..)
+    , WrapInsertion (..)
     , PreConversion (..)
     , defaultShakespeareSettings
     , shakespeare
@@ -18,15 +21,23 @@ module Text.Shakespeare
     , shakespeareFromString
     , shakespeareUsedIdentifiers
     , RenderUrl
-    , VarType
+    , VarType (..)
     , Deref
+    , Parser
 
 #ifdef TEST_EXPORT
     , preFilter
 #endif
+      -- * Internal
+      -- can we remove this?
+    , shakespeareRuntime
+    , pack'
     ) where
 
-import Text.ParserCombinators.Parsec hiding (Line)
+import Data.List (intersperse)
+import Data.Char (isAlphaNum, isSpace)
+import Text.ParserCombinators.Parsec hiding (Line, parse, Parser)
+import Text.Parsec.Prim (modifyState, Parsec)
 import Language.Haskell.TH.Quote (QuasiQuoter (..))
 import Language.Haskell.TH (appE)
 import Language.Haskell.TH.Syntax
@@ -40,8 +51,30 @@ import qualified Data.Text as TS
 import qualified Data.Text.Lazy as TL
 import Text.Shakespeare.Base
 
+import Filesystem (getModified)
+import Filesystem.Path.CurrentOS (decodeString)
+import Data.Time (UTCTime)
+import Data.IORef
+import qualified Data.Map as M
+import GHC.Generics (Generic)
+import Data.Typeable (Typeable)
+import Data.Data (Data)
+
 -- for pre conversion
-import System.Process (readProcess)
+import System.Process (readProcessWithExitCode)
+import System.Exit (ExitCode(..))
+
+#if !MIN_VERSION_base(4,5,0)
+(<>) :: Monoid m => m -> m -> m
+(<>) = mappend
+{-# INLINE (<>) #-}
+#endif
+
+-- | A parser with a user state of [String]
+type Parser = Parsec String [String]
+-- | run a parser with a user state of [String]
+parse ::  GenParser tok [a1] a -> SourceName -> [tok] -> Either ParseError a
+parse p = runParser p []
 
 -- move to Shakespeare.Base?
 readFileQ :: FilePath -> Q String
@@ -51,7 +84,7 @@ readFileQ fp = qRunIO $ readFileUtf8 fp
 readFileUtf8 :: FilePath -> IO String
 readFileUtf8 fp = fmap TL.unpack $ readUtf8File fp
 
--- | The Coffeescript language compiles down to Javascript.
+-- | Coffeescript, TypeScript, and other languages compiles down to Javascript.
 -- Previously we waited until the very end, at the rendering stage to perform this compilation.
 -- Lets call is a post-conversion
 -- This had the advantage that all Haskell values were inserted first:
@@ -61,19 +94,34 @@ readFileUtf8 fp = fmap TL.unpack $ readUtf8File fp
 -- The down-side is the template must be compiled down to Javascript during every request.
 -- If instead we do a pre-conversion to compile down to Javascript,
 -- we only need to perform the compilation once.
+--
+-- The problem then is the insertion of Haskell values: we need a hole for
+-- them. This can be done with variables known to the language.
 -- During the pre-conversion we first modify all Haskell insertions
--- so that they will be ignored by the Coffeescript compiler (backticks).
--- So %{var} is change to `%{var}` using the preEscapeBegin and preEscapeEnd.
+-- So #{a} is change to shakespeare_var_a
+-- Then we can place the Haskell values in a function wrapper that exposes
+-- those variables: (function(shakespeare_var_a){ ... shakespeare_var_a ...})
+-- TypeScript can compile that, and then we tack an application of the
+-- Haskell values onto the result: (#{a})
+--
 -- preEscapeIgnoreBalanced is used to not insert backtacks for variable already inside strings or backticks.
 -- coffeescript will happily ignore the interpolations, and backticks would not be treated as escaping in that context.
 -- preEscapeIgnoreLine was added to ignore comments (which in Coffeescript begin with a '#')
 
 data PreConvert = PreConvert
     { preConvert :: PreConversion
-    , preEscapeBegin :: String
-    , preEscapeEnd   :: String
     , preEscapeIgnoreBalanced :: [Char]
     , preEscapeIgnoreLine :: [Char]
+    , wrapInsertion :: Maybe WrapInsertion
+    }
+
+data WrapInsertion = WrapInsertion {
+      wrapInsertionIndent     :: Maybe String
+    , wrapInsertionStartBegin :: String
+    , wrapInsertionSeparator  :: String
+    , wrapInsertionStartClose :: String
+    , wrapInsertionEnd :: String
+    , wrapInsertionAddParens :: Bool
     }
 
 data PreConversion = ReadProcess String [String]
@@ -107,8 +155,12 @@ defaultShakespeareSettings = ShakespeareSettings {
 }
 
 instance Lift PreConvert where
-    lift (PreConvert convert begin end ignore comment) =
-        [|PreConvert $(lift convert) $(lift begin) $(lift end) $(lift ignore) $(lift comment)|]
+    lift (PreConvert convert ignore comment wrapInsertion) =
+        [|PreConvert $(lift convert) $(lift ignore) $(lift comment) $(lift wrapInsertion)|]
+
+instance Lift WrapInsertion where
+    lift (WrapInsertion indent sb sep sc e wp) =
+        [|WrapInsertion $(lift indent) $(lift sb) $(lift sep) $(lift sc) $(lift e) $(lift wp)|]
 
 instance Lift PreConversion where
     lift (ReadProcess command args) =
@@ -178,20 +230,88 @@ parseContents = many1 . parseContent
         parseChar' = ContentRaw `fmap` many1 (noneOf [varChar, urlChar, intChar])
 
 
-preFilter :: ShakespeareSettings -> String -> IO String
-preFilter ShakespeareSettings {..} s = 
-    case preConversion of
-      Nothing -> return s
-      Just pre@(PreConvert convert _ _ _ _) ->
-        let parsed = mconcat $ eShowErrors $ parse (parseConvert pre) s s
-        in  case convert of
-              Id -> return parsed
-              ReadProcess command args ->
-                readProcess command args parsed
+-- | calls 'error' when there is stderr or exit code failure
+readProcessError :: FilePath -> [String] -> String
+                 -> Maybe FilePath -- ^ for error reporting
+                 -> IO String
+readProcessError cmd args input mfp = do
+  (ex, output, err) <- readProcessWithExitCode cmd args input
+  case ex of
+   ExitSuccess   ->
+     case err of
+       [] -> return output
+       msg -> error $ "stderr received during readProcess:" ++ displayCmd ++ "\n\n" ++ msg
+   ExitFailure r ->
+    error $ "exit code " ++ show r ++ " from readProcess: " ++ displayCmd ++ "\n\n"
+      ++ "stderr:\n" ++ err
   where
-    parseConvert PreConvert {..} = many1 $ choice $
-        map (try . escapedParse) preEscapeIgnoreBalanced ++
-        [mainParser]
+    displayCmd = cmd ++ ' ':unwords (map show args) ++
+        case mfp of
+          Nothing -> ""
+          Just fp -> ' ':fp
+
+preFilter :: Maybe FilePath -- ^ for error reporting
+          -> ShakespeareSettings
+          -> String
+          -> IO String
+preFilter mfp ShakespeareSettings {..} template =
+    case preConversion of
+      Nothing -> return template
+      Just pre@(PreConvert convert _ _ mWrapI) ->
+        if all isSpace template then return template else
+          let (groups, rvars) = eShowErrors $ parse
+                                  (parseConvertWrapInsertion mWrapI pre)
+                                  template
+                                  template
+              vars = reverse rvars
+              parsed = mconcat groups
+              withVars = (addVars mWrapI vars parsed)
+          in  applyVars mWrapI vars `fmap` case convert of
+                  Id -> return withVars
+                  ReadProcess command args ->
+                    readProcessError command args withVars mfp
+  where
+    addIndent :: Maybe String -> String -> String
+    addIndent Nothing str = str
+    addIndent (Just indent) str = mapLines (\line -> indent <> line) str
+      where
+        mapLines f = unlines . map f . lines
+
+    shakespeare_prefix = "shakespeare_var_"
+    shakespeare_var_conversion ('@':'?':'{':str) = shakespeare_var_conversion ('@':'{':str)
+    shakespeare_var_conversion (_:'{':str) = shakespeare_prefix <> filter isAlphaNum (init str)
+    shakespeare_var_conversion err = error $ "did not expect: " <> err
+
+    applyVars _      [] str = str
+    applyVars Nothing _ str = str
+    applyVars (Just WrapInsertion {..}) vars str =
+         (if wrapInsertionAddParens then "(" else "")
+      <> removeTrailingSemiColon
+      <> (if wrapInsertionAddParens then ")" else "")
+      <> "("
+      <> mconcat (intersperse ", " vars)
+      <> ");\n"
+        where 
+          removeTrailingSemiColon = reverse $
+             dropWhile (\c -> c == ';' || isSpace c) (reverse str)
+
+    addVars _      [] str = str
+    addVars Nothing _ str = str
+    addVars (Just WrapInsertion {..}) vars str =
+         wrapInsertionStartBegin
+      <> mconcat (intersperse wrapInsertionSeparator $ map shakespeare_var_conversion vars)
+      <> wrapInsertionStartClose
+      <> addIndent wrapInsertionIndent str
+      <> wrapInsertionEnd
+
+    parseConvertWrapInsertion Nothing = parseConvert id
+    parseConvertWrapInsertion (Just _) = parseConvert shakespeare_var_conversion
+
+    parseConvert varConvert PreConvert {..} = do
+        str <- many1 $ choice $
+          map (try . escapedParse) preEscapeIgnoreBalanced ++ [mainParser]
+        st <- getState
+        return (str, st)
 
       where
         escapedParse ignoreC = do
@@ -207,8 +327,8 @@ preFilter ShakespeareSettings {..} s =
             parseCommentLine preEscapeIgnoreLine <|>
             parseChar' preEscapeIgnoreLine preEscapeIgnoreBalanced
 
-        escape str = preEscapeBegin ++ str ++ preEscapeEnd
-        escapeRight = either id escape
+        recordRight (Left str)  = return str
+        recordRight (Right str) = modifyState (\vars -> str:vars) >> return (varConvert str)
 
         newLine = "\r\n"
         parseCommentLine cs = do
@@ -216,9 +336,10 @@ preFilter ShakespeareSettings {..} s =
           comment <- many $ noneOf newLine
           return $ begin : comment
 
-        parseVar' = escapeRight `fmap` parseVarString varChar
-        parseUrl' = escapeRight `fmap` parseUrlString urlChar '?'
-        parseInt' = escapeRight `fmap` parseIntString intChar
+        parseVar' :: (Parsec String [String]) String
+        parseVar' = recordRight =<< parseVarString varChar
+        parseUrl' = recordRight =<< parseUrlString urlChar '?'
+        parseInt' = recordRight =<< parseIntString intChar
         parseChar' comments ignores =
             many1 (noneOf ([varChar, urlChar, intChar] ++ comments ++ ignores))
 
@@ -240,17 +361,17 @@ contentsToShakespeare rs a = do
         _   -> do
               mc <- [|mconcat|]
               return $ mc `AppE` ListE c
-    fmap (maybe id AppE $ modifyFinalValue rs) $
+    fmap (maybe id AppE $ modifyFinalValue rs) $ return $
         if justVarInterpolation rs
-            then return compiledTemplate
-            else return $ LamE [VarP r] compiledTemplate
+            then compiledTemplate
+            else LamE [VarP r] compiledTemplate
       where
         contentToBuilder :: Name -> Content -> Q Exp
         contentToBuilder _ (ContentRaw s') = do
             ts <- [|fromText . pack'|]
             return $ wrap rs `AppE` (ts `AppE` LitE (StringL s'))
         contentToBuilder _ (ContentVar d) =
-            return $ wrap rs `AppE` (toBuilder rs `AppE` derefToExp [] d)
+            return (toBuilder rs `AppE` derefToExp [] d)
         contentToBuilder r (ContentUrl d) = do
             ts <- [|fromText|]
             return $ wrap rs `AppE` (ts `AppE` (VarE r `AppE` derefToExp [] d `AppE` ListE []))
@@ -266,17 +387,22 @@ shakespeare r = QuasiQuoter { quoteExp = shakespeareFromString r }
 
 shakespeareFromString :: ShakespeareSettings -> String -> Q Exp
 shakespeareFromString r str = do
-    s <- qRunIO $ preFilter r str
+    s <- qRunIO $ preFilter Nothing r $
+#ifdef WINDOWS
+          filter (/='\r')
+#endif
+          str
     contentsToShakespeare r $ contentFromString r s
 
 shakespeareFile :: ShakespeareSettings -> FilePath -> Q Exp
-shakespeareFile r fp = do
+shakespeareFile r fp =
 #ifdef GHC_7_4
-    qAddDependentFile fp
+    qAddDependentFile fp >>
 #endif
-    readFileQ fp >>= shakespeareFromString r
+        readFileQ fp >>= shakespeareFromString r
 
 data VarType = VTPlain | VTUrl | VTUrlParam | VTMixin
+    deriving (Show, Eq, Ord, Enum, Bounded, Typeable, Data, Generic)
 
 getVars :: Content -> [(Deref, VarType)]
 getVars ContentRaw{} = []
@@ -287,7 +413,7 @@ getVars (ContentMix d) = [(d, VTMixin)]
 
 data VarExp url = EPlain Builder
                 | EUrl url
-                | EUrlParam (url, [(TS.Text, TS.Text)])
+                | EUrlParam (url, QueryParameters)
                 | EMixin (Shakespeare url)
 
 -- | Determine which identifiers are used by the given template, useful for
@@ -295,16 +421,30 @@ data VarExp url = EPlain Builder
 shakespeareUsedIdentifiers :: ShakespeareSettings -> String -> [(Deref, VarType)]
 shakespeareUsedIdentifiers settings = concatMap getVars . contentFromString settings
 
+type MTime = UTCTime
+
+{-# NOINLINE reloadMapRef #-}
+reloadMapRef :: IORef (M.Map FilePath (MTime, [Content]))
+reloadMapRef = unsafePerformIO $ newIORef M.empty
+
+lookupReloadMap :: FilePath -> IO (Maybe (MTime, [Content]))
+lookupReloadMap fp = do
+  reloads <- readIORef reloadMapRef
+  return $ M.lookup fp reloads
+
+insertReloadMap :: FilePath -> (MTime, [Content]) -> IO [Content]
+insertReloadMap fp (mt, content) = atomicModifyIORef reloadMapRef
+  (\reloadMap -> (M.insert fp (mt, content) reloadMap, content))
+
 shakespeareFileReload :: ShakespeareSettings -> FilePath -> Q Exp
-shakespeareFileReload rs fp = do
+shakespeareFileReload settings fp = do
     str <- readFileQ fp
-    s <- qRunIO $ preFilter rs str
-    let b = shakespeareUsedIdentifiers rs s
+    s <- qRunIO $ preFilter (Just fp) settings str
+    let b = shakespeareUsedIdentifiers settings s
     c <- mapM vtToExp b
-    rt <- [|shakespeareRuntime|]
-    wrap' <- [|\x -> $(return $ wrap rs) . x|]
-    r' <- lift rs
-    return $ wrap' `AppE` (rt `AppE` r' `AppE` (LitE $ StringL fp) `AppE` ListE c)
+    rt <- [|shakespeareRuntime settings fp|]
+    wrap' <- [|\x -> $(return $ wrap settings) . x|]
+    return $ wrap' `AppE` (rt `AppE` ListE c)
   where
     vtToExp :: (Deref, VarType) -> Q Exp
     vtToExp (d, vt) = do
@@ -313,18 +453,31 @@ shakespeareFileReload rs fp = do
         return $ TupE [d', c' `AppE` derefToExp [] d]
       where
         c :: VarType -> Q Exp
-        c VTPlain = [|EPlain . $(return $ toBuilder rs)|]
+        c VTPlain = [|EPlain . $(return $
+          InfixE (Just $ unwrap settings) (VarE '(.)) (Just $ toBuilder settings))|]
         c VTUrl = [|EUrl|]
         c VTUrlParam = [|EUrlParam|]
-        c VTMixin = [|\x -> EMixin $ \r -> $(return $ unwrap rs) $ x r|]
+        c VTMixin = [|\x -> EMixin $ \r -> $(return $ unwrap settings) $ x r|]
+
 
 
 shakespeareRuntime :: ShakespeareSettings -> FilePath -> [(Deref, VarExp url)] -> Shakespeare url
-shakespeareRuntime rs fp cd render' = unsafePerformIO $ do
-    str <- readFileUtf8 fp
-    s <- preFilter rs str
-    return $ mconcat $ map go $ contentFromString rs s
+shakespeareRuntime settings fp cd render' = unsafePerformIO $ do
+    mtime <- qRunIO $ getModified $ decodeString fp
+    mdata <- lookupReloadMap fp
+    case mdata of
+      Just (lastMtime, lastContents) ->
+        if mtime == lastMtime then return $ go' lastContents
+          else fmap go' $ newContent mtime
+      Nothing -> fmap go' $ newContent mtime
   where
+    newContent mtime = do
+        str <- readFileUtf8 fp
+        s <- preFilter (Just fp) settings str
+        insertReloadMap fp (mtime, contentFromString settings s)
+
+    go' = mconcat . map go
+
     go :: Content -> Builder
     go (ContentRaw s) = fromText $ TS.pack s
     go (ContentVar d) =
